package raytracer.utilities;

//Copyright (C) Helen Hu 2011.
//based on C++ code from Ray Tracing from the Ground Up, by Kevin Suffern 
//This Java code is for non-commercial purposes only.
//This Java code is licensed under the GNU General Public License Version 2.
//See the file COPYING.txt for the full license.

public class Vector3D {
	public double	x, y, z;
	
	// ---------------------------------------------------------- default constructor
	public Vector3D() {
		x = y = z = 0.0;
	}

	// ---------------------------------------------------------- constructor

	public Vector3D(double a) {
		x = y = z = a;
	}

	// ---------------------------------------------------------- constructor

	public Vector3D(double a, double b, double c)	 
	{
		x = a;
		y = b;
		z = c;
	}

	// ---------------------------------------------------------- copy constructor

	public Vector3D(Vector3D vector) {
		x = vector.x;
		y = vector.y;
		z = vector.z;
	}


	// ---------------------------------------------------------- constructor
	// constructs a vector from a normal

	public Vector3D(Normal n)	
	{
		x = n.x;
		y = n.y;
		z = n.z;
	}

	// ---------------------------------------------------------- constructor
	// constructs a vector from a point
	// this is used in the ConcaveHemisphere hit functions

	public Vector3D(Point3D p)	 
	{
		x = p.x;
		y = p.y;
		z = p.z;
	}

	// ---------------------------------------------------------- assignment operator

	public void set (Vector3D rhs) {
		if (this != rhs) {
			x = rhs.x; y = rhs.y; z = rhs.z;
		}
	}

	// ----------------------------------------------------------- assignment operator
	// assign a Normal to a vector

	public void set (Normal rhs) {
		x = rhs.x; y = rhs.y; z = rhs.z;
	}

	// ---------------------------------------------------------- assignment operator 
	// assign a point to a vector
	
	public void set (Point3D rhs) {
		x = rhs.x; y = rhs.y; z = rhs.z;
	}

	// ------------------------------------------------------------------------ unary minus
	// this does not change the current vector
	// this allows ShadeRec objects to be declared as constant arguments in many shading
	// functions that reverse the direction of a ray that's stored in the ShadeRec object

	public Vector3D minus() {
		return new Vector3D(-x, -y, -z);    
	}

	// ----------------------------------------------------------  length
	// length of the vector

	public double length() {
		return (Math.sqrt(x * x + y * y + z * z));
	}

	// ---------------------------------------------------------------------  len_squared
	// the square of the length

	public double len_squared() {	
		return (x * x + y * y + z * z);
	}


	// ----------------------------------------------------------------------- operator*
	// multiplication by a double on the right

	public Vector3D multiply (double a) {	
		return new Vector3D(x * a, y * a, z * a);	
	}

	// ----------------------------------------------------------------------- operator/
	// division by a double

	public Vector3D divide (double a) {	
		return new Vector3D(x / a, y / a, z / a);	
	}


	// ----------------------------------------------------------------------- operator+
	// addition

	public Vector3D add (Vector3D v) {
		return new Vector3D(x + v.x, y + v.y, z + v.z);
	}


	// ----------------------------------------------------------------------- operator-
	// subtraction

	public Vector3D subtract (Vector3D v) {
		return new Vector3D(x - v.x, y - v.y, z - v.z);
	}


	// ----------------------------------------------------------------------- operator*
	// dot product

	public double dot (Vector3D v) {
		return (x * v.x + y * v.y + z * v.z);
	} 

	// ----------------------------------------------------------------------- operator*
	// dot product

	public double dot (Normal v) {
		return (x * v.x + y * v.y + z * v.z);
	} 

	// ----------------------------------------------------------------------- operator^
	// cross product

	public Vector3D cross (Vector3D v)  {
		return new Vector3D(y * v.z - z * v.y, z * v.x - x * v.z, x * v.y - y * v.x);
	}

	// ----------------------------------------------------------------------- operator^
	// cross product

	public Vector3D cross (Normal v)  {
		return new Vector3D(y * v.z - z * v.y, z * v.x - x * v.z, x * v.y - y * v.x);
	}

	// ---------------------------------------------------------------------  operator+=
	// compound addition

	public void plusEqual(Vector3D v) {
		x += v.x; y += v.y; z += v.z;
	}

	// ----------------------------------------------------------  normalize
	// converts the vector to a unit vector

	public void normalize() {	
		double length = Math.sqrt(x * x + y * y + z * z);
		x /= length; y /= length; z /= length;
	}
	
	// ----------------------------------------------------------  hat
	// converts the vector to a unit vector and returns the vector

	public Vector3D hat() {	
		double length = Math.sqrt(x * x + y * y + z * z);
		x /= length; y /= length; z /= length;
		return this;
	}
	
	public boolean equals(Object obj) {
		if (obj == null || this.getClass() != obj.getClass()) {
			return false;
		}
		else {
			Vector3D other = (Vector3D)obj;
			return (this.x==other.x && this.y==other.y && this.z==other.z);
		}
	}
	
	public String toString() {
		return "Vector3D: (" + x + "," + y + "," + z + ")";
	}
	
	// static methods
	// non-member function

	// ----------------------------------------------------------------------- operator*
	// multiplication by a double on the left

	public Vector3D multiply (double a, Vector3D v) {
		return new Vector3D(a * v.x, a * v.y, a * v.z);	
	}

	// ----------------------------------------------------------  operator* 
	// multiplication by a matrix on the left

	public static Vector3D multiply (Matrix mat, Vector3D v) {
		return new Vector3D(mat.m[0][0] * v.x + mat.m[0][1] * v.y + mat.m[0][2] * v.z,
						mat.m[1][0] * v.x + mat.m[1][1] * v.y + mat.m[1][2] * v.z,
						mat.m[2][0] * v.x + mat.m[2][1] * v.y + mat.m[2][2] * v.z);
	}

	
	
}
