package raytracer.world;
import raytracer.utilities.*;
import raytracer.geometricObjects.*;
import raytracer.tracers.*;
import raytracer.scenes.*;
import java.util.Vector;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.WritableRaster;

//Copyright (C) Helen Hu 2011.
//based on C++ code from Ray Tracing from the Ground Up, by Kevin Suffern 
//This Java code is for non-commercial purposes only.
//This Java code is licensed under the GNU General Public License Version 2.
//See the file COPYING.txt for the full license.


//This file contains the declaration of the class World
//The World class does not have a copy constructor or an assignment operator, for the following reasons:

//1 	There's no need to copy construct or assign the World
//2 	We wouldn't want to do this anyway, because the world can contain an arbitrary amount of data
//3 	These operations wouldn't work because the world is self-referencing:
//	 	the Tracer base class contains a pointer to the world. If we wrote a correct copy constructor for the 
//	  	Tracer class, the World copy constructor would call itself recursively until we ran out of memory. 

public class World {
	public ViewPlane				vp;
	public RGBColor					background_color;
	public Tracer					tracer;
	public Sphere 					sphere;		// for Chapter 3 only
	public Vector<GeometricObject>	objects;		

	private Graphics				graphics;   // for displaying to RayTracerGUI

	// for saving image
	public BufferedImage image;
	private WritableRaster raster;
	private ColorModel model;

	private static final int ROWS = 300, COLS = 300;


	// -------------------------------------------------------------------- default constructor

	public World(){
		vp = new ViewPlane();
		background_color = RGBColor.black;
		tracer= null;
		sphere = new Sphere();
		objects = new Vector<GeometricObject>();
	}

	public void setGraphics(Graphics gr) {
		graphics = gr;
	}

	// ------------------------------------------------------------------ add_object

	public void add_object(GeometricObject object) {  
		objects.add(object);	
	}

	// ------------------------------------------------------------------ build

	/*
	 * change this method if you want to build different items
	 */
	public int[] build() {
		vp.set_hres(COLS);
		vp.set_vres(ROWS);
		vp.set_pixel_size(1.f);
		vp.set_gamma(1.f);

		Chapter3.buildSingleSphere(this);
		//		Chapter3.buildMultipleObjects(this);
		//		Chapter3.buildBBCoverPic(this);

		int[] resolution = new int[2];
		resolution[0] = vp.hres;
		resolution[1] = vp.vres;

		// for saving the image to a file
		image = new BufferedImage(vp.hres,vp.vres, BufferedImage.TYPE_INT_RGB);
		raster = image.getRaster();
		model = image.getColorModel();

		return resolution;
	}

	//------------------------------------------------------------------ render_scene

	// This uses orthographic viewing along the zw axis

	public void render_scene()  {

		RGBColor	pixel_color;	 	
		Ray			ray 	= new Ray();					
		double		zw		= 100.0;			// hardwired in
		double		x, y;

		ray.d = new Vector3D(0, 0, -1);

		for (int r = 0; r < vp.vres; r++)			// up
			for (int c = 0; c <= vp.hres; c++) {	// across 					
				x = vp.s * (c - 0.5 * (vp.hres - 1.0));
				y = vp.s * (r - 0.5 * (vp.hres - 1.0));
				ray.o = new Point3D(x, y, zw);
				pixel_color = tracer.trace_ray(ray);
				display_pixel(r, c, pixel_color);
			}	
	}  

	// ------------------------------------------------------------------ clamp

	public RGBColor max_to_one(RGBColor c) {
		float max_value = Math.max(c.r, Math.max(c.g, c.b));

		if (max_value > 1.0)
			return new RGBColor(c.divide(max_value));
		else
			return new RGBColor(c);
	}


	// ------------------------------------------------------------------ clamp_to_color
	// Set color to red if any component is greater than one

	public RGBColor clamp_to_color(RGBColor raw_color) {
		RGBColor c = new RGBColor(raw_color);

		if (raw_color.r > 1.0 || raw_color.g > 1.0 || raw_color.b > 1.0) {
			c.r = 1.0f; c.g = 0.0f; c.b = 0.0f;
		}

		return (c);
	}


	// ---------------------------------------------------------------------------display_pixel

	// raw_color is the pixel color computed by the ray tracer
	// its RGB floating point components can be arbitrarily large
	// mapped_color has all components in the range [0, 1], but still floating point
	// display color has integer components for computer display
	// the Mac's components are in the range [0, 65535]
	// a PC's components will probably be in the range [0, 255]
	// the system-dependent code is in the function convert_to_display_color
	// the function SetCPixel is a Mac OS function


	public void display_pixel(int row, int column, RGBColor raw_color) {
		RGBColor mapped_color;

		if (vp.show_out_of_gamut)
			mapped_color = clamp_to_color(raw_color);
		else
			mapped_color = max_to_one(raw_color);

		if (vp.gamma != 1.0)
			mapped_color = mapped_color.powc(vp.inv_gamma);

		//have to start from max y coordinate to convert to screen coordinates
		int x = column;
		int y = vp.vres - row - 1;

		// for Java window
		Color color = new Color((int)(mapped_color.r * 255),
				(int)(mapped_color.g * 255),
				(int)(mapped_color.b * 255));

		graphics.setColor(color);

		graphics.fillRect(x, y, 1, 1);

		// for saving image to file
//		int rgb = color.getRGB();		
//		Object colorData = model.getDataElements(rgb, null);
//		raster.setDataElements(x, y, colorData);
		
	}


	// ----------------------------------------------------------------------------- hit_bare_bones_objects

	public ShadeRec hit_bare_bones_objects(Ray ray) {
		ShadeRec	sr = new ShadeRec(this); 
		double		t; 			
		double		tmin 			= Constants.kHugeValue;
		int 		num_objects 	= objects.size();

		for (int j = 0; j < num_objects; j++) {
			t = objects.get(j).hit(ray, sr);
			if ((0 < t) && (t < tmin)) {
				sr.hit_an_object	= true;
				tmin 				= t; 
				sr.color			= objects.get(j).get_color(); 
			}
		}

		return (sr);   
	}


	//------------------------------------------------------------------ delete_objects

	// Deletes the objects in the objects array, and erases the array.
	// The objects array still exists, because it's an automatic variable, but it's empty 

	public void delete_objects() {
		objects.clear();
	}

}

